/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
  * © 2016 OX Software GmbH, Germany. info@open-xchange.com
 *
 * @author Kai Ahrens <kai.ahrens@open-xchange.com>
 * @author Carsten Driesner <carsten.driesner@open-xchange.com>
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/editframework/app/rtconnection', [
    'io.ox/core/event',
    'io.ox/core/uuids',
    'io.ox/core/tk/sessionrestore',
    'io.ox/realtime/groups',
    'io.ox/office/tk/utils',
    'io.ox/office/tk/utils/logger',
    'io.ox/office/tk/object/triggerobject',
    'io.ox/office/tk/object/timermixin',
    'io.ox/office/editframework/utils/editconfig'
], function (Events, UUID, SessionRestore, RTGroups, Utils, Logger, TriggerObject, TimerMixin, Config) {

    'use strict';

    var // names of internal events for the real-time connection state
        INTERNAL_EVENTS = 'online offline reset error:notMember',

        // names of runtime requests expecting an answer
        REQUEST_EVENTS = 'addresource getdocument',

        // names of runtime push messages
        PUSH_EVENTS = 'update flusherrorinfo hangup preparelosingeditrights declinedacquireeditrights acceptedacquireeditrights infomessage renameandreload alive',

        // repeated time to check for pending backend ACKs
        CHECK_PENDING_INTERVAL = 30000,

        EVENT_NOUPDATEFROMSERVER = 'noUpdateFromServer';

    // private static functions ===============================================

    /**
     * Encodes file properties of the passed file descriptor to a resource
     * identifier that can be used to initialize a real-time connection.
     *
     * @param {Object} file
     *  A file descriptor object which contains file specific properties, like
     *  'id' and 'folder_id'.
     *
     * @returns {String}
     *  A string which references a file and complies to the resource id
     *  encoding scheme; or null, if the file descriptor doesn't describe a
     *  valid file.
     */
    function encodeFileAsResourceId(file) {
        return encodeURIComponent(file['com.openexchange.realtime.resourceID'] || file.id);
    }

    // class RTConnection =====================================================

    /**
     * Represents the connection to the real-time framework, used to send
     * message to and receive message from the server.
     *
     * Triggers the following events:
     * - 'update': Update notifications containing actions created by remote
     *  clients, and state information for the application (read-only, name and
     *  identifier of the current editor, etc.).
     *
     * @extends TriggerObject
     * @extends TimerMixin
     *
     * @param {EditApplication} app
     *  The application containing this operations queue.
     *
     * @param {Object} [initOptions]
     *  Optional parameters:
     *  @param {Number} [initOptions.realTimeDelay=700]
     *      The duration in milliseconds the realtime framework will collect
     *      messages before sending them to the server.
     */
    function RTConnection(app, initOptions) {

        var // self reference
            self = this,

            // file descriptor of the edited document
            file = app.getFileDescriptor(),

            // rt channel id
            channelId = app.getDocumentType(),

            // create the core real-time group instance for the document
            rtGroup = RTGroups.getGroup('synthetic.' + channelId + '://operations/' + encodeFileAsResourceId(file)),

            // private event hub for internal event handling
            eventHub = new Events(),

            // duration for the realtime framework collecting messages before sending
            realTimeDelay = Utils.getIntegerOption(initOptions, 'realTimeDelay', 700, 0),

            traceActions = {},

            // action messages sent but not yet ACK
            sendActionWaitForACK = [],

            // joining the group
            joining = false,

            // rejoining is possible
            rejoinPossible = false,

            // old rtid of the user for restore support
            oldrtId = SessionRestore.state('documents.rtid'),

            // stores the current detected offline state
            offline = false;

        // store rtid to retrieve it with a possible browser reload
        SessionRestore.state('documents.rtid', RTGroups.rtId);

        // base constructor ---------------------------------------------------

        TriggerObject.call(this);
        TimerMixin.call(this);

        // private methods ----------------------------------------------------

        /**
         * Log the data
         */
        function debugLogUpdateNotification(data) {
            var // the operations array (may be missing completely)
                actions = Utils.getArrayOption(data, 'actions', []),
                // editing user id
                editUserId = Utils.getStringOption(data, 'editUserId', ''),
                // server osn
                serverOSN = Utils.getIntegerOption(data, 'serverOSN', -1),
                // has errors
                hasErrors = Utils.getBooleanOption(data, 'hasErrors', false),
                // server osn
                activeClients = Utils.getArrayOption(data, 'activeUsers', []).length,
                // write protected
                writeProtected = Utils.getBooleanOption(data, 'writeProtected', false),
                // locking user
                lockedByUser = Utils.getStringOption(data, 'lockedByUser', ''),
                // locking user id
                lockedByUserId = Utils.getIntegerOption(data, 'lockedByUserId', -1),
                // output string
                output = '';

            // output content of update notification
            output = 'editUserId: ' + editUserId + '; ';
            output += 'server osn: ' + serverOSN + '; ';
            output += 'active users: ' + activeClients + '; ';
            output += 'hasErrors: ' + hasErrors + '; ';
            output += 'writeProtected: ' + writeProtected + '; ';
            output += 'locked by user: ' + lockedByUser + '; ';
            output += 'locked by id: ' + lockedByUserId;
            Utils.log(output);

            // check that all array elements are objects
            if (_.all(actions, _.isObject)) {
                try {
                    Utils.log('Last remote action received by RT-Update: ' + JSON.stringify(actions[actions.length - 1]));
                } catch (ex) {
                    // do nothing
                }
            }
        }

        /**
         * Register events for specific RT actions.
         * The event handler 'eventName' is triggered, when
         * an RT event with the action name 'eventName' is received.
         * The triggered eventName will be equal to the 'actionName'.
         * The eventHandler will be called with the received payload data
         */
        function registerRTEvents(eventNames) {
            _.each(eventNames.split(/\s+/), function (eventName) {
                rtGroup.on('receive', function (event, stanza) {
                    _.each(stanza.getAll(channelId, eventName), function (payload) {
                        RTConnection.log('RTConnection.eventHub.trigger (' + channelId + '), ' + eventName);
                        eventHub.trigger(eventName, payload.data);
                    });
                });
            });
        }

        /**
         * Determines whether a certain action should be traced on its way
         * through the realtime backend.
         *
         * @returns {Boolean}
         *  Whether to trace the stanzas for the specified action.
         */
        function shouldTrace(action) {
            return traceActions[action.toLowerCase()];
        }

        /**
         * Checks the pending actions backend ACKs for their pending
         * time.
         */
        function checkPendingAcks() {
            var // current time
                now = Date.now(),
                // state if noUpdateFromServer must be triggered
                triggerNoUpdate = false;

            if (!offline) {
                _.each(sendActionWaitForACK, function (entry) {
                    var // the time elapsed since the action has been sent
                        timeSpan = now - entry.timeStamp;

                    if (timeSpan > CHECK_PENDING_INTERVAL) {
                        triggerNoUpdate = true;
                    }
                });

                if (triggerNoUpdate) {
                    self.trigger(EVENT_NOUPDATEFROMSERVER);
                }
            }
        }

        /**
         * Removes the actions which received a ack from an backend-side update
         * message.
         *
         * {Object} data
         *  Contains the "update" data received from the backend.
         */
        function removeAckActions(data) {
            var // the data of the ack action messages
                acks = Utils.getArrayOption(data, 'acks', null),
                // a map for faster access
                ackMap = {};

            if (acks && (acks.length > 0)) {
                // create map of the ack'ed waiting actions
                _.each(acks, function (ack) {
                    ackMap[ack] = true;
                });

                // filter out all actions which have a map entry
                sendActionWaitForACK = _.filter(sendActionWaitForACK, function (waitAction) {
                    return _.isUndefined(ackMap[waitAction.id]);
                });
            }
        }

        /**
         * Sends a message to the connected RT object, waiting for the
         * acknowledge to this call.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved if the acknowledge for this request
         *  will have arrived, or rejected after a time out otherwise.
         */
        function send(action, element, data) {

            var payloads = [{ element: 'action', data: action }],
                defer = null;

            if (_.isString(element) && !_.isUndefined(data)) {
                payloads.push({
                    namespace: channelId,
                    element: element,
                    data: data,
                    verbatim: true
                });
            }

            defer = rtGroup.send({
                element: 'message',
                payloads: payloads,
                trace: shouldTrace(action),
                bufferinterval: realTimeDelay
            });

            // Time out detection now uses the deferred to trigger an optional timeout handler
            defer.fail(function () {
                RTConnection.error('Send message with RT failed, payloads: ' + payloads.toString());
                // RT calls us even after connection, rtgroup are destroyed and references are cleared.
                if (!self.destroyed) { self.trigger('timeout'); }
            });

            return defer.promise();
        }

        /**
         * Sends a message request to the connected RT object, returning a
         * promise object that will be resolved, if the result data to this
         * request has arrived.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved if the result for this request has
         *  arrived, or rejected otherwise.
         */
        function sendQuery(action, element, data) {
            var payloads = [{ element: 'action', data: action }],
                defer;

            if (_.isString(element) && !_.isUndefined(data)) {
                payloads.push({
                    namespace: channelId,
                    element: element,
                    data: data,
                    verbatim: true
                });
            }
            // Add necessary payload to provide the real-time id for
            // correct rights management on the backend
            payloads.push({
                namespace: 'office',
                element: 'rtdata',
                data: { rtid: RTGroups.rtId, session: ox.session },
                verbatim: true
            });

            defer = rtGroup.query({
                element: 'message',
                payloads: payloads,
                trace: shouldTrace(action),
                bufferinterval: realTimeDelay
            });

            // Time out detection now uses the deferred to trigger an optional timeout handler
            defer.fail(function () {
                RTConnection.error('Send query message with RT failed, payloads: ' + payloads.toString());
                // RT calls us even after connection, rtgroup are destroyed and references are cleared.
                if (!self.destroyed) {
                    self.trigger('timeout');
                }
            });

            return defer.promise();
        }

        /**
         * Handles the response from the RT low-level code and extract the
         * data part from, providing it as result via the promise.
         *
         * @param {jQuery.Promise} promise
         *  The original promise provided by the RT low-level code.
         *
         * @param {Number} [timeout]
         *  If specified, the delay time in milliseconds, after the promise
         *  returned by this method will be rejected, if the passed original
         *  promise is still pending.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved if the result for the original
         *  request has been received, or rejected otherwise.
         */
        function handleResponse(promise, timeout) {

            var // the Deferred object to be returned
                def = $.Deferred();

            promise.then(function (response) {
                handleResolvedResponse(def, response);
            }, function (response) {
                handleRejectedResponse(def, response);
            });

            // prevent endless waiting
            if (_.isNumber(timeout) && (timeout > 0)) {
                var timer = self.executeDelayed(function () {
                    if (def.state() === 'pending') {
                        RTConnection.log('Timeout reached for RT request, timeout=' + timeout + 'ms.');
                        def.reject({ cause: 'timeout' });
                    }
                }, timeout, 'RTConnection: handleResponse');
                def.always(function () { timer.abort(); timer = null; });
            }

            return def.promise();
        }

        /**
         * Handles the response from the RT low-level code and adds timeout
         *
         * @param {jQuery.Promise} promise
         *  The original promise provided by the RT low-level code.
         *
         * @param {Number} [timeout]
         *  If specified, the delay time in milliseconds, after the Promise
         *  returned by this method will be rejected, if the passed original
         *  promise is still pending.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved if the result for the original
         *  request has been received, or rejected otherwise.
         */
        function handleGenericResponse(promise, timeout) {

            var // the Deferred object to be returned
                def = $.Deferred();

            promise.then(function (data) { def.resolve(data); }, function (data) { def.reject(data); });

            // prevent endless waiting
            if (_.isNumber(timeout) && (timeout > 0)) {
                self.executeDelayed(function () { def.reject({ cause: 'timeout' }); }, timeout, 'RTConnection: handleGenericResponse');
            }

            return def.promise();
        }

        /**
         * Handles the resolved response of the RT low-level code and
         * tries to extract the data port from it.
         *
         * @param {jQuery.Deferred} def
         *  The deferred to be resolved with the extracted data.
         *
         * @param {Object} response
         *  The original response provided by the RT low-level code.
         */
        function handleResolvedResponse(def, response) {
            var payloads = Utils.getArrayOption(response, 'payloads', []);

            if (self.destroyed) {
                // reject deferred after destroy
                def.reject({ cause: 'closing' });
            }
            if (payloads.length > 0) {
                // Extract the data and provide it
                def.resolve(payloads[0].data);
            } else {
                // Response from server is not in the expected format, therefore
                // we reject the deferred here.
                def.reject(response);
            }
        }

        /**
         * Handles the rejected response of the RT low-level code and
         * tries to extract the data port from it.
         *
         * @param {jQuery.Deferred} def
         *  The deferred to be rejected with the extracted data.
         *
         * @param {Object} response
         *  The original response provided by the RT low-level code.
         */
        function handleRejectedResponse(def, response) {
            var payloads = Utils.getArrayOption(response, 'payloads', []);

            if (payloads.length > 0) {
                // Extract data even for the rejected case
                def.reject(payloads[0].data);
            } else {
                // If we don't find a payload just provide the original response
                def.reject(response);
            }
        }

        /**
         * Handles the resolved response of the RT low-level join request
         * and makes checks for certain aspects (errors).
         *
         * @param {jQuery.Deferred} def
         *
         * @param {Object} response
         *  The original response from a join-request provided by the RT
         *  low-level code.
         */
        function handleJoinResponse(def, response) {
            var // the payloads of the join response
                payloads = Utils.getArrayOption(response, 'payloads', []),
                // the data to resolve/reject the deferred
                data = null,
                // do we found an error
                hasError = false;

            if (self.destroyed) {
                // reject deferred after destroy
                def.reject({ cause: 'closing' });
            }

            if (payloads.length > 0) {
                // Extract the answer payload from the backend answer, look
                // for possible errors sent by RT via payloads, too. Should be
                // handled by RT-core (see #39911).
                _.find(payloads, function (payload) {
                    if (_.isObject(payload) && _.isObject(payload.data) && _.isString(payload.element)) {
                        switch (payload.element) {
                            case 'error': hasError = true; data = { error: payload.data }; break;
                            case 'getactions': data = payload.data; break;
                            default: /* ignore payload */ break;
                        }
                    }
                    return hasError;
                });

                data = data || payloads[0].data; // as fallback
                if (hasError) { def.reject(data); } else { def.resolve(data); }
            } else {
                // we must receive at least one payload - otherwise we
                // assume an error
                def.reject(response);
            }
        }

        /**
         * Handles the rejected response of the RT low-level join request
         * and makes checks for certain aspects (errors). In special cases
         * a re-join is started to re-enroll the client.
         *
         * @param {jQuery.Deferred} def
         *
         * @param {Object} response
         *  The original response from a join-request provided by the RT
         *  low-level code.
         */
        function handleRejectedJoinRequest(def, response, retryHandler) {
            var // do we found error data
                hasError = false;

            if (self.destroyed) {
                // reject deferred after destroy
                def.reject({ cause: 'closing' });
            }

            // we assume that we receive the error data as a property in the response
            if (_.isObject(response) && _.isObject(response.error)) {
                hasError = true;
            }

            // check error code to decide, if we can try to re-join
            if (hasError && response.error.code === 1007) {
                // try to re-join in case of code 1007 (STATE_MISSING)
                handleRejoin(def, response, retryHandler);
            } else {
                joining = false;
                def.reject(response);
            }
        }

        /**
         * Handles the re-join of the client, in case a re-enroll
         * is necessary.
         *
         * @param {Function} retryHandler
         *  A handler which is responsible to start a re-join of the client.
         *  Must be valid function.
         */
        function handleRejoin(def, response, retryHandler) {
            // We have to wait for possible notifications from the
            // low-level RT part to determine, if a retry is possible.
            self.executeDelayed(function () {
                // now check the states and decide, if we can try to
                // start a reconnect using a possible retry handler
                if (rejoinPossible) {
                    rejoinPossible = false;
                    retryHandler.call(this, def);
                    joining = false;
                } else {
                    handleJoinResponse(def, response);
                    joining = false;
                }
            }, 1000, 'RTConnection: joinGroup');
        }

        /**
         * Join a group using the provided options and set the asynchronous
         * state. An optional retryHandler is called, if there is a chance
         * to retry the joining.
         *
         * @param {Object} connectOptions
         *  Contains properties to be sent while joining to the
         *  document real-time connection.
         *
         * @param {jQuery.Deferred} def
         *  The deferred to be resolved/rejected, if joining completed/failed.
         *
         * @param {Function} retryHandler
         *  A function that is called, if a retry is possible.
         */
        function joinGroup(connectOptions, def, retryHandler) {

            joining = true;
            rtGroup.join(connectOptions).then(function (response) {
                var payloads = Utils.getArrayOption(response, 'payloads', []);

                if (self.destroyed) {
                    // reject deferred after destroy
                    def.reject({ cause: 'closing' });
                    joining = false;
                } else if (payloads.length > 0) {
                    handleJoinResponse(def, response);
                    joining = false;
                } else {
                    // We encountered an illegal state for our join request.
                    if (_.isFunction(retryHandler)) {
                        // try to re-join if handler is available
                        handleRejoin(def, response, retryHandler);
                    } else {
                        // without retry handler just process the original
                        // response
                        handleJoinResponse(def, response);
                        joining = false;
                    }
                }
            }, function (response) {
                handleRejectedJoinRequest(def, response, retryHandler);
            });
        }

        // methods ------------------------------------------------------------

        /**
         * Connects the real-time group and sends the initial request for the
         * document import operations.
         *
         * @param {Object} connectData
         *  Contains properties to be sent while joining to the
         *  document real-time connection.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved with the getActions results.
         */
        this.connect = function (connectData) {

            var connectDeferred = $.Deferred(),
                // optional connect data
                connectOptions = { expectWelcomeMessage: true },
                // connect data
                sendConnectData = connectData || {};

            // ATTENTION:
            // We have to send an additional payload with context information for the join.
            // Due to sharing we need to provide folder & file-id to enable the backend to
            // load the document & check the permissions in the context of every user.
            // The session alone is not enough anymore. Please be aware that this is now
            // MANDATORY and changing this can break the whole load/save process!!
            _.extend(connectOptions, { additionalPayloads: [{ element: 'connectData', namespace: channelId, data: _.extend(sendConnectData, { oldrtId: oldrtId, folder_id: file.folder_id, id: file.id, session: ox.session }), verbatim: true }] });

            if (!rtGroup.isRTWorking()) {
                // the server doesn't support real-time communication, must be
                // incorrect configured.
                return $.Deferred().reject({ cause: 'bad server component' });
            }

            RTConnection.log('RTConnection.connect called');

            // Handler to be called, if we detected a rejoin scenario - possible
            // if a re-enroll is needed by the low-level RT framework.
            function rejoinHandler(def) {
                // call join again, but this time without retry handler
                joinGroup(connectOptions, def, null);
            }

            // joins the group and provides the retry handler
            joinGroup(connectOptions, connectDeferred, rejoinHandler);

            // Remember *connect* deferred to be able to reject them if user
            // closes document while we are still connecting to the realtime group!
            return this.createAbortablePromise(connectDeferred);
        };

        /**
         * Send the given operations to the connected RT object
         * for further distribution to other connected clients
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the ACK of the internal send
         *  arrives.
         */
        this.sendActions = function (actions) {
            var // wrap the action array into an object and add a
                // unique id for later identification and a time stamp to enable
                // us to use a time out mechanism
                actionData = { id: UUID.randomUUID(), timeStamp: Date.now(), action: actions };

            RTConnection.log('RTConnection.sendActions, id: ' + actionData.id + ', timeStamp: ' + actionData.timeStamp + ', actions: ' + actions);

            // push action data to our waiting queue
            sendActionWaitForACK.push(actionData);

            return send('applyactions', 'actions', actionData);
        };

        /**
         * Sends user data update to the server. This user update data will
         * broadcasted to all collaborating clients.
         *
         * @param {String} action
         *  Action name string to be handled by the server.
         *
         * @param {Object} userData
         *  An object containing relevant additional user data on the document.
         *
         * @returns {jQuery.Promise}
         */
        this.updateUserData = Config.SHOW_REMOTE_SELECTIONS ? function (action, userData) {
            RTConnection.log('RTConnection.updateUserData called', userData);
            return send(action, 'state', userData);
        } : function () { return $.when(); };

        /**
         * Sends a real-time request to the server and waits for a response.
         *
         * @param {String} action
         *  The identifier of the action sent to the server.
         *
         * @param {Number} [timeout]
         *  If specified, the delay time in milliseconds, after the Promise
         *  returned by this method will be rejected, if the query is still
         *  pending.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved with the answer of the server
         *  request, or rejected on error or timeout.
         */
        this.sendQuery = function (action, timeout) {
            RTConnection.log('RTConnection.sendQuery(' + action + ')');
            return handleResponse(sendQuery(action), timeout);
        };

        /**
         * Initiates flushing of the document using a synchronous request.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the answer for flushDocument
         *  from the server arrives.
         */
        this.flushDocument = function () {
            RTConnection.log('RTConnection.flushDocument called');
            return handleGenericResponse(sendQuery('flushdocument'), 20000);
        };

        /**
         * Send our wish to acquire the edit rights to the connected RT object
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the ACK of the internal send
         *  arrives.
         */
        this.acquireEditRights = function () {
            RTConnection.log('RTConnection.acquireedit called');
            return send('acquireedit', 'state', { osn: app.getModel().getOperationStateNumber() });
        };

        /**
         * Acknowledge to the server that our preparation to lose the
         * edit rights are completed.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the ACK of the internal send
         *  arrives.
         */
        this.canLoseEditRights = function () {
            RTConnection.log('RTConnection.canloseeditrights called');
            return send('canloseeditrights', 'state', { osn: app.getModel().getOperationStateNumber() });
        };

        /**
         * Sends client log message to the server so it can be written to the
         * server side log files for later debugging.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the ACK of the internal send
         *  arrives.
         */
        this.sendLogMessage = function (data) {
            RTConnection.log('RTConnection.sendLogMessage called');
            return send('logMessage', 'state', { message: data });
        };

        /**
         * Sends a sync request to the server, which will answer with a update
         * notification. This can be used to synchronize client and server
         * after an off-/online scenario.
         */
        this.sync = function () {
            RTConnection.log('RTConnection.sync called');
            return send('sync', 'state', { osn: app.getModel().getOperationStateNumber() });
        };

        /**
         * Sends a "alive" answer to the server, which extends the time
         * for client to make necessary changes until a "edit rights" switch will
         * be forced due to time-out.
         *
         * @return {jQuery.Promise}
         *  A promise that will be resolved when the ACK of the internal send
         *  arrives.
         */
        this.alive = function () {
            RTConnection.log('RTConnection.alive called');
            return send('alive', 'state', { osn: app.getModel().getOperationStateNumber() });
        };

        /**
         * Close the document at the connected RT object
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved with the closing document status of
         *  the request.
         */
        this.closeDocument = function (recentFileData) {
            var leaveOptions = { expectSignOffMessage: true };

            RTConnection.log('RTConnection.closeDocument called');
            if (recentFileData) {
                _.extend(leaveOptions, { additionalPayloads: [{ element: 'recentfile', namespace: channelId, data: recentFileData, verbatim: true }] });
            }
            return handleResponse(rtGroup.leave(leaveOptions), 20000);
        };

        /**
         * Provides the unique real-time id used by this connection.
         *
         * @returns {String}
         *  The unique real-time id used by this connection.
         */
        this.getUuid = function () {
            return RTGroups.rtId;
        };

        /**
         * Retrieves the pending actions, not processed by the backend. These
         * actions can be provided to a restoreDocument() function to try to
         * restore the document.
         *
         * @returns {Object}
         *  The result object contains the property "actions" which is an Array
         *  with all pending operations (without server-side ack).
         */
        this.getPendingActions = function () {
            var // result of the merge of the pending operations
                pendingActions = [];

            if ((_.isArray(sendActionWaitForACK)) && (sendActionWaitForACK.length > 0)) {
                // the sendActionWaitForACK contains the actions with additional data:
                // {
                //    id: uuid,
                //    timeStamp: time of putting it into the queue,
                //    action: the ops array
                // }
                _.each(sendActionWaitForACK, function (element) {
                    // extract the operations from our wrapper object and
                    // put it into the result array
                    pendingActions = pendingActions.concat(element.action);
                });
            }

            return { actions: pendingActions };
        };

        /**
         * Determines, if we have pending actions which have not been acknowledge by
         * the OX Documents backend.
         *
         * @returns {Boolean} TRUE, if there are pending actions, otherwise FALSE.
         */
        this.hasPendingActions = function () {
            return ((_.isArray(sendActionWaitForACK)) && (sendActionWaitForACK.length > 0));
        };

        // initialization -----------------------------------------------------

        RTConnection.log('RTConnection initialization');
        this.repeatDelayed(checkPendingAcks, CHECK_PENDING_INTERVAL, 'RTConnection: checkPendingAcks');

        // read traceable actions from page URL
        if (_.url.hash('office:trace')) {
            _.each(_.url.hash('office:trace').split(/\s*,\s*/), function (action) {
                traceActions[action.toLowerCase()] = true;
            });
        }

        // register the internal event hub for request events and custom runtime events
        registerRTEvents(REQUEST_EVENTS + ' ' + PUSH_EVENTS);

        // forward runtime realtime events to own listeners
        rtGroup.on(INTERNAL_EVENTS, function (event, data) {
            var // trigger event
                triggerEvent = true;

            // special handling for reset event
            if (event.type === 'reset') {
                if (joining) {
                    // If we are joining, reset is an indicator that a re-enroll
                    // has been done. In that case we have to try to join again.
                    // This must be limited to error code 1007 which means that
                    // we have to re-enroll. A reset after we received a reset with
                    // data.code 1007 is also normal, ignore that, too.
                    // See #40096, the error structure changed
                    if ((!rejoinPossible && (data && data.code === 1007)) || rejoinPossible) {
                        if (!rejoinPossible) {
                            rejoinPossible = true;
                        }
                        triggerEvent = false;
                    }
                }
            } else if (event.type === 'offline') {
                offline = true;
            } else if (event.type === 'online') {
                offline = false;
            }

            if (triggerEvent) {
                self.trigger(event.type);
            }
        });

        // forward OX Documents runtime events to own listeners
        eventHub.on(PUSH_EVENTS, function (event, data) {
            if (event.type === 'update') {
                removeAckActions(data);
                if (RTConnection.debug) {
                    debugLogUpdateNotification(data);
                }
            }
            self.trigger(event.type, data);
        });

        // disconnect from the RT object, no further calls should be made at this object from now on
        this.registerDestructor(function () {

            // destroy realtime group
            try {
                rtGroup.destroy();
                rtGroup = null;
                RTConnection.log('RTConnection.destroy(): RT group destroyed');
            } catch (ex) {
                Utils.exception(ex);
            }

            // destroy private event hub
            eventHub.destroy();
            eventHub = null;

            // remove references
            file = traceActions = null;

            // remove actions from pending queue waiting for backend ACKs
            sendActionWaitForACK = null;
        });

    } // class RTConnection

    // global initialization ==================================================

    // whether to show console output of this RTConnection class
    RTConnection.debug = Config.DEBUG && Config.getDebugState('realtimedebugging', 'office:log-realtime');

    // extend statically with logging methods
    Logger.extend(RTConnection, { enable: RTConnection.debug, prefix: 'RT' });

    // enable logging for real-time core
    require(['io.ox/realtime/rt']).done(function (CoreRT) {
        CoreRT.debug = RTConnection.debug;
    });

    // exports ================================================================

    // derive this class from class TriggerObject
    return TriggerObject.extend({ constructor: RTConnection });

});
